use std::path::{PathBuf, Path, StripPrefixError};
use roxy_core::error::Error;

#[derive(Debug, Clone)]
pub(crate) struct FilePath<'a, P: AsRef<Path>> {
    pub input: PathBuf,
    pub root_dir: PathBuf,
    pub output: &'a P,
}

impl<'a, P: AsRef<Path> + 'a> FilePath<'a, P> {
    pub fn new(input: &'a P, output: &'a P) -> Self {
        Self {
            input: Self::make_recursive(input),
            root_dir: Self::strip_wildcards(input),
            output,
        }
    }

    fn make_recursive(path: &'a P) -> PathBuf {
        path.as_ref().join("**/*")
    }

    fn has_no_wildcard<S: AsRef<str>>(path: &S) -> bool {
        !path.as_ref().contains("*")
    }

    fn strip_wildcards<P2: AsRef<Path> + ?Sized>(path: &'a P2) -> PathBuf {
        path.as_ref()
            .ancestors()
            .map(Path::to_str)
            .flatten()
            .find(Self::has_no_wildcard)
            .map_or_else(|| PathBuf::new(), PathBuf::from)
    }

    pub fn to_output<P2: AsRef<Path>>(&self, value: &'a P2) -> Result<PathBuf, Error> {
        value
            .as_ref()
            .strip_prefix(&self.root_dir)
            .map(|path| self.output.as_ref().join(path))
            .map_err(Error::from)
    }

    pub fn strip_root<P2: AsRef<Path>>(&self, value: &'a P2) -> Result<&Path, StripPrefixError> {
        value.as_ref().strip_prefix(&self.root_dir)
    }
}



