pub mod config;
pub mod context;
mod file_path;

use config::Config;
use context::Context;
use file_path::FilePath;
use roxy_core::error::Error;

use roxy_markdown_parser::MarkdownParser;
use roxy_markdown_tera_rewriter::{MarkdownTeraPreformatter, MarkdownTeraRewriter};
use roxy_syntect::SyntectParser;
use roxy_tera_parser::{TeraParser, TeraParserOptions};

use clap::Parser as Clap;
use std::{
    ffi,
    fs::{self, File},
    io::{BufReader, Read},
    path::{Path, PathBuf},
};
use toml::Table;

use glob::glob;
use roxy_core::roxy::{Parser, Roxy};

fn handle_err<E: std::error::Error + 'static>(err: E) -> Error {
    Error::new(err.to_string(), err)
}

#[derive(Clap)]
#[command(name = "Roxy")]
#[command(author = "KitsuneCafe")]
#[command(version = "2.0")]
#[command(about = "A very small static site generator", long_about = None)]
pub struct Options {
    pub input: String,
    pub output: String,
}

fn get_files<P: AsRef<Path> + std::fmt::Debug>(path: &P) -> Result<Vec<PathBuf>, Error> {
    let path = path
        .as_ref()
        .to_str()
        .ok_or_else(|| Error::from(format!("{path:?} is not a valid path.")))?;

    let files: Vec<PathBuf> = glob(path)
        .map_err(handle_err)?
        .filter_map(|x| x.ok())
        .filter(|f| Path::is_file(f))
        .collect();

    Ok(files)
}

fn load_config(path: &Path) -> Config {
    fs::read_to_string(path).map_or_else(
        |_| Config::default(),
        |f| {
            toml::from_str::<Config>(f.as_str())
                .unwrap()
                .merge(Config::default())
        },
    )
}

fn context_from_meta_files<'a, T: AsRef<Path>>(
    files: &Vec<&PathBuf>,
    file_path: &'a FilePath<T>,
) -> Result<Context, Error> {
    let mut context = Context::new();

    for path in files {
        let mut buf = Vec::new();

        let mut file = File::open(path).map(BufReader::new)?;
        file.read_to_end(&mut buf)?;
        let mut str = String::from_utf8(buf).map_err(handle_err)?;
        let toml: Table = toml::from_str(&mut str).map_err(handle_err)?;

        context.insert(
            &file_path.strip_root(path)?,
            &tera::to_value(toml).map_err(handle_err)?,
        );
    }

    Ok(context)
}

fn copy_static<T: AsRef<Path>>(files: &Vec<&PathBuf>, file_path: &FilePath<T>) -> Result<(), Error> {
    for file in files {
        let output = file_path.to_output(file)?;
        fs::create_dir_all(output.parent().unwrap())?;
        fs::copy(file, file_path.to_output(file)?)?;
    }

    Ok(())
}

const DEFAULT_THEME: &'static str = "base16-ocean.dark";
const CONTENT_EXT: [&'static str; 4] = ["md", "tera", "html", "htm"];

fn main() -> Result<(), Box<dyn std::error::Error>> {
    let opts = Options::parse();

    let file_path = FilePath::new(&opts.input, &opts.output);

    let config_path = file_path.input.with_file_name("config.toml");
    let config = load_config(&config_path);

    let files = get_files(&file_path.input)?;
    let (meta, files): (Vec<&PathBuf>, Vec<&PathBuf>) =
        files.iter().partition(|f| f.extension().unwrap() == "toml");

    let (content, files): (Vec<&PathBuf>, Vec<&PathBuf>) = files
        .iter()
        .partition(|f| {
            let ext = f.extension().and_then(ffi::OsStr::to_str).unwrap();
            CONTENT_EXT.contains(&ext)
        });

    let mut context = context_from_meta_files(&meta, &file_path)?;

    let theme = config.theme.unwrap_or(DEFAULT_THEME.to_string());

    for file in content {
        let file_name = file.with_extension("html");
        let output_path = file_path.to_output(&file_name)?;

        let mut parser = Parser::new();
        let mut preformatter = MarkdownTeraPreformatter::new();
        parser.push(&mut preformatter);

        let mut syntect = SyntectParser::new(theme.as_str());
        parser.push(&mut syntect);

        let mut md = MarkdownParser::new();
        parser.push(&mut md);

        let mut rewriter = MarkdownTeraRewriter::new();
        parser.push(&mut rewriter);

        let file_name = file.with_extension("html");

        if let Ok(path) = &file_path.strip_root(&file_name) {
            if let Some(current_context) = context.get(path) {
                context.insert(&"this", &current_context.clone());
            }
        }

        let mut tera = tera::Tera::default();
        let mut html = TeraParser::new(&mut tera, TeraParserOptions::default());
        html.add_context(context.to_inner());
        parser.push(&mut html);

        Roxy::process_file(&file, &output_path, &mut parser).unwrap();
    }

    copy_static(&files, &file_path)?;

    Ok(())
}
